@@ -1,9 +1,22 @@
+# -*- coding: utf-8 -*-
 
 import py
 from py._code.code import FormattedExcinfo, ReprExceptionInfo
 queue = py.builtin._tryimport('queue', 'Queue')
 
 failsonjython = py.test.mark.xfail("sys.platform.startswith('java')")
+from test_source import astonly
+
+try:
+    import importlib
+except ImportError:
+    invalidate_import_caches = None
+else:
+    invalidate_import_caches = getattr(importlib, "invalidate_caches", None)
+
+import pytest
+pytest_version_info = tuple(map(int, pytest.__version__.split(".")[:3]))
+pytest25 = pytest.mark.skipif(pytest_version_info < (2,5), reason="requires pytest-2.5")
 
 class TWMock:
     def __init__(self):
@@ -84,6 +97,7 @@ class TestTraceback_f_g_h:
         assert s.startswith("def f():")
         assert s.endswith("raise ValueError")
 
+    @astonly
     @failsonjython
     def test_traceback_entry_getsource_in_construct(self):
         source = py.code.Source("""\
@@ -101,7 +115,7 @@ class TestTraceback_f_g_h:
             print (tb[-1].getsource())
             s = str(tb[-1].getsource())
             assert s.startswith("def xyz():\n    try:")
-            assert s.endswith("except somenoname:")
+            assert s.strip().endswith("except somenoname:")
 
     def test_traceback_cut(self):
         co = py.code.Code(f)
@@ -115,7 +129,7 @@ class TestTraceback_f_g_h:
     def test_traceback_cut_excludepath(self, testdir):
         p = testdir.makepyfile("def f(): raise ValueError")
         excinfo = py.test.raises(ValueError, "p.pyimport().f()")
-        basedir = py._pydir
+        basedir = py.path.local(py.test.__file__).dirpath()
         newtraceback = excinfo.traceback.cut(excludepath=basedir)
         for x in newtraceback:
             if hasattr(x, 'path'):
@@ -154,6 +168,25 @@ class TestTraceback_f_g_h:
         recindex = traceback.recursionindex()
         assert recindex is None
 
+    def test_traceback_messy_recursion(self):
+        #XXX: simplified locally testable version
+        decorator = py.test.importorskip('decorator').decorator
+
+        def log(f, *k, **kw):
+            print('%s %s' % (k, kw))
+            f(*k, **kw)
+        log = decorator(log)
+
+        def fail():
+            raise ValueError('')
+
+        fail = log(log(fail))
+
+        excinfo = py.test.raises(ValueError, fail)
+        assert excinfo.traceback.recursionindex() is None
+
+
+
     def test_traceback_getcrashentry(self):
         def i():
             __tracebackhide__ = True
@@ -211,6 +244,15 @@ def test_excinfo_exconly():
     assert msg.startswith('ValueError')
     assert msg.endswith("world")
 
+@pytest25
+def test_excinfo_exconly_unicode_AssertionError():
+    val = py.builtin._totext('£€', 'utf-8')
+    def fail():
+        raise AssertionError(val)
+    excinfo = py.test.raises(Exception, fail)
+    msg = excinfo.exconly(tryshort=True)
+    assert msg == 'AssertionError: ' + val
+
 def test_excinfo_repr():
     excinfo = py.test.raises(ValueError, h)
     s = repr(excinfo)
@@ -238,6 +280,23 @@ def test_excinfo_no_sourcecode():
     else:
         assert s == "  File '<string>':1 in <module>\n  ???\n"
 
+def test_excinfo_no_python_sourcecode(tmpdir):
+    #XXX: simplified locally testable version
+    tmpdir.join('test.txt').write("{{ h()}}:")
+
+    jinja2 = py.test.importorskip('jinja2')
+    loader = jinja2.FileSystemLoader(str(tmpdir))
+    env = jinja2.Environment(loader=loader)
+    template = env.get_template('test.txt')
+    excinfo = py.test.raises(ValueError,
+                             template.render, h=h)
+    for item in excinfo.traceback:
+        print(item) #XXX: for some reason jinja.Template.render is printed in full
+        item.source # shouldnt fail
+        if item.path.basename == 'test.txt':
+            assert str(item.source) == '{{ h()}}:'
+
+
 def test_entrysource_Queue_example():
     try:
         queue.Queue().get(timeout=0.001)
@@ -268,6 +327,8 @@ class TestFormattedExcinfo:
             modpath = tmpdir.join("mod.py")
             tmpdir.ensure("__init__.py")
             modpath.write(source)
+            if invalidate_import_caches is not None:
+                invalidate_import_caches()
             return modpath.pyimport()
         return importasmod
 
@@ -296,15 +357,11 @@ class TestFormattedExcinfo:
     def test_repr_source_excinfo(self):
         """ check if indentation is right """
         pr = FormattedExcinfo()
-        py.code.patch_builtins()
-        try:
-            excinfo = self.excinfo_from_exec("""
+        excinfo = self.excinfo_from_exec("""
                 def f():
                     assert 0
                 f()
-            """)
-        finally:
-            py.code.unpatch_builtins()
+        """)
         pr = FormattedExcinfo()
         source = pr._getentrysource(excinfo.traceback[-1])
         lines = pr.get_source(source, 1, excinfo)
@@ -342,6 +399,8 @@ raise ValueError()
         pr = FormattedExcinfo()
 
         class FakeCode(object):
+            class raw:
+                co_filename = '?'
             path = '?'
             firstlineno = 5
 
@@ -352,12 +411,16 @@ raise ValueError()
         class FakeFrame(object):
             code = FakeCode()
             f_locals = {}
+            f_globals = {}
 
         class FakeTracebackEntry(py.code.Traceback.Entry):
             def __init__(self, tb):
-                self.frame = FakeFrame()
                 self.lineno = 5+3
 
+            @property
+            def frame(self):
+                return FakeFrame()
+
         class Traceback(py.code.Traceback):
             Entry = FakeTracebackEntry
 
@@ -388,7 +451,7 @@ raise ValueError()
 
     def test_repr_local(self):
         p = FormattedExcinfo(showlocals=True)
-        loc = {'y': 5, 'z': 7, 'x': 3, '__builtins__': {}} # __builtins__}
+        loc = {'y': 5, 'z': 7, 'x': 3, '@x': 2, '__builtins__': {}}
         reprlocals = p.repr_locals(loc)
         assert reprlocals.lines
         assert reprlocals.lines[0] == '__builtins__ = <builtins>'
@@ -451,6 +514,27 @@ raise ValueError()
         assert tw.lines[1] == "x = 5, y = 13"
         assert tw.lines[2] == "z = " + repr('z' * 120)
 
+    def test_repr_tracebackentry_lines_var_kw_args(self, importasmod):
+        mod = importasmod("""
+            def func1(x, *y, **z):
+                raise ValueError("hello\\nworld")
+        """)
+        excinfo = py.test.raises(ValueError, mod.func1, 'a', 'b', c='d')
+        excinfo.traceback = excinfo.traceback.filter()
+        entry = excinfo.traceback[-1]
+        p = FormattedExcinfo(funcargs=True)
+        reprfuncargs = p.repr_args(entry)
+        assert reprfuncargs.args[0] == ('x', repr('a'))
+        assert reprfuncargs.args[1] == ('y', repr(('b',)))
+        assert reprfuncargs.args[2] == ('z', repr({'c': 'd'}))
+
+        p = FormattedExcinfo(funcargs=True)
+        repr_entry = p.repr_traceback_entry(entry)
+        assert repr_entry.reprfuncargs.args == reprfuncargs.args
+        tw = TWMock()
+        repr_entry.toterminal(tw)
+        assert tw.lines[0] == "x = 'a', y = ('b',), z = {'c': 'd'}"
+
     def test_repr_tracebackentry_short(self, importasmod):
         mod = importasmod("""
             def func1():
@@ -553,6 +637,22 @@ raise ValueError()
             assert repr.reprcrash.path.endswith("mod.py")
             assert repr.reprcrash.message == "ValueError: 0"
 
+    def test_repr_traceback_with_invalid_cwd(self, importasmod, monkeypatch):
+        mod = importasmod("""
+            def f(x):
+                raise ValueError(x)
+            def entry():
+                f(0)
+        """)
+        excinfo = py.test.raises(ValueError, mod.entry)
+
+        p = FormattedExcinfo()
+        def raiseos():
+            raise OSError(2)
+        monkeypatch.setattr(py.std.os, 'getcwd', raiseos)
+        assert p._makepath(__file__) == __file__
+        reprtb = p.repr_traceback(excinfo)
+
     def test_repr_excinfo_addouterr(self, importasmod):
         mod = importasmod("""
             def entry():
@@ -605,11 +705,7 @@ raise ValueError()
                 x = 1
                 assert x == 2
         """)
-        py.code.patch_builtins(assertion=True)
-        try:
-            excinfo = py.test.raises(AssertionError, mod.somefunc)
-        finally:
-            py.code.unpatch_builtins(assertion=True)
+        excinfo = py.test.raises(AssertionError, mod.somefunc)
 
         p = FormattedExcinfo()
         reprentry = p.repr_traceback_entry(excinfo.traceback[-1], excinfo)
@@ -631,6 +727,14 @@ raise ValueError()
                 assert isinstance(repr, ReprExceptionInfo)
                 assert repr.reprtraceback.style == style
 
+    def test_reprexcinfo_unicode(self):
+        from py._code.code import TerminalRepr
+        class MyRepr(TerminalRepr):
+            def toterminal(self, tw):
+                tw.line(py.builtin._totext("я", "utf-8"))
+        x = py.builtin._totext(MyRepr())
+        assert x == py.builtin._totext("я", "utf-8")
+
     def test_toterminal_long(self, importasmod):
         mod = importasmod("""
             def g(x):
@@ -707,9 +811,11 @@ raise ValueError()
             assert 0
         """)
         repr = excinfo.getrepr(style='native')
-        assert repr.startswith('Traceback (most recent call last):\n  File')
-        assert repr.endswith('\nAssertionError: assert 0\n')
-        assert 'exec (source.compile())' in repr
+        assert "assert 0" in str(repr.reprcrash)
+        s = str(repr)
+        assert s.startswith('Traceback (most recent call last):\n  File')
+        assert s.endswith('\nAssertionError: assert 0')
+        assert 'exec (source.compile())' in s
         # python 2.4 fails to get the source line for the assert
         if py.std.sys.version_info >= (2, 5):
-            assert repr.count('assert 0') == 2
+            assert s.count('assert 0') == 2
