@@ -1,3 +1,5 @@
+# -*- coding: utf-8 -*-
+
 from __future__ import generators
 import py
 import sys
@@ -60,6 +62,11 @@ class TestSafeRepr:
         except Exception:
             py.test.fail("saferepr failed for newstyle class")
 
+    def test_unicode(self):
+        val = py.builtin._totext('£€', 'utf-8')
+        reprval = py.builtin._totext("'£€'", 'utf-8')
+        assert saferepr(val) == reprval
+
 def test_unicode_handling():
     value = py.builtin._totext('\xc4\x85\xc4\x87\n', 'utf-8').encode('utf8')
     def f():
@@ -69,18 +76,3 @@ def test_unicode_handling():
     if sys.version_info[0] < 3:
         u = unicode(excinfo)
 
-def test_unicode_or_repr():
-    from py._code.code import unicode_or_repr
-    assert unicode_or_repr('hello') == "hello"
-    if sys.version_info[0] < 3:
-        s = unicode_or_repr('\xf6\xc4\x85')
-    else:
-        s = eval("unicode_or_repr(b'\\f6\\xc4\\x85')")
-    assert 'print-error' in s
-    assert 'c4' in s
-    class A:
-        def __repr__(self):
-            raise ValueError()
-    s = unicode_or_repr(A())
-    assert 'print-error' in s
-    assert 'ValueError' in s
