@@ -1,5 +1,9 @@
+# -*- coding: utf-8 -*-
+
+from __future__ import with_statement
 import py
-import sys
+import pytest
+import os, sys
 from py.path import local
 import common
 
@@ -14,7 +18,7 @@ skiponwin32 = py.test.mark.skipif(
 
 def pytest_funcarg__path1(request):
     def setup():
-        path1 = request.config.mktemp("path1")
+        path1 = request.getfuncargvalue("tmpdir")
         common.setuptestfs(path1)
         return path1
     def teardown(path1):
@@ -74,13 +78,40 @@ class TestLocalPath(common.CommonFSTests):
     def test_initialize_curdir(self):
         assert str(local()) == py.std.os.getcwd()
 
+    @skiponwin32
+    def test_chdir_gone(self, path1):
+        p = path1.ensure("dir_to_be_removed", dir=1)
+        p.chdir()
+        p.remove()
+        pytest.raises(py.error.ENOENT, py.path.local)
+        assert path1.chdir() is None
+        assert os.getcwd() == str(path1)
+
+    def test_as_cwd(self, path1):
+        dir = path1.ensure("subdir", dir=1)
+        old = py.path.local()
+        with dir.as_cwd() as x:
+            assert x == old
+            assert py.path.local() == dir
+        assert os.getcwd() == str(old)
+
+    def test_as_cwd_exception(self, path1):
+        old = py.path.local()
+        dir = path1.ensure("subdir", dir=1)
+        with pytest.raises(ValueError):
+            with dir.as_cwd():
+                raise ValueError()
+        assert old == py.path.local()
+
     def test_initialize_reldir(self, path1):
-        old = path1.chdir()
-        try:
+        with path1.as_cwd():
             p = local('samplefile')
             assert p.check()
-        finally:
-            old.chdir()
+
+    @pytest.mark.xfail("sys.version_info < (2,6) and sys.platform == 'win32'")
+    def test_tilde_expansion(self):
+        p = py.path.local("~", expanduser=True)
+        assert p == os.path.expanduser("~")
 
     def test_eq_with_strings(self, path1):
         path1 = path1.join('sampledir')
@@ -91,6 +122,31 @@ class TestLocalPath(common.CommonFSTests):
         assert path3 != path2
         assert path2 != path3
 
+    def test_eq_with_none(self, path1):
+        assert path1 != None
+
+    def test_gt_with_strings(self, path1):
+        path2 = path1.join('sampledir')
+        path3 = str(path1.join("ttt"))
+        assert path3 > path2
+        assert path2 < path3
+        assert path2 < "ttt"
+        assert "ttt" > path2
+        path4 = path1.join("aaa")
+        l = [path2, path4,path3]
+        assert sorted(l) == [path4, path2, path3]
+
+    def test_open_and_ensure(self, path1):
+        p = path1.join("sub1", "sub2", "file")
+        with p.open("w", ensure=1) as f:
+            f.write("hello")
+        assert p.read() == "hello"
+
+    def test_write_and_ensure(self, path1):
+        p = path1.join("sub1", "sub2", "file")
+        p.write("hello", ensure=1)
+        assert p.read() == "hello"
+
     @py.test.mark.multi(bin=(False, True))
     def test_dump(self, tmpdir, bin):
         path = tmpdir.join("dumpfile%s" % int(bin))
@@ -178,7 +234,7 @@ class TestLocalPath(common.CommonFSTests):
     def test_init_from_path(self, tmpdir):
         l = local()
         l2 = local(l)
-        assert l2 is l
+        assert l2 == l
 
         wc = py.path.svnwc('.')
         l3 = local(wc)
@@ -212,22 +268,36 @@ class TestLocalPath(common.CommonFSTests):
         # check that breadth comes last
         assert l[2] == p3
 
+    def test_visit_rec_fnmatch(self, tmpdir):
+        p1 = tmpdir.ensure("a","123")
+        p2 = tmpdir.ensure(".b","345")
+        l = list(tmpdir.visit("???", rec="[!.]*"))
+        assert len(l) == 1
+        # check that breadth comes last
+        assert l[0] == p1
+
+    def test_fnmatch_file_abspath(self, tmpdir):
+        b = tmpdir.join("a", "b")
+        assert b.fnmatch(os.sep.join("ab"))
+        pattern = os.sep.join([str(tmpdir), "*", "b"])
+        assert b.fnmatch(pattern)
+
+    def test_sysfind(self):
+        name = sys.platform == "win32" and "cmd" or "test"
+        x = py.path.local.sysfind(name)
+        assert x.check(file=1)
+        assert py.path.local.sysfind('jaksdkasldqwe') is None
+        assert py.path.local.sysfind(name, paths=[]) is None
+        x2 = py.path.local.sysfind(name, paths=[x.dirpath()])
+        assert x2 == x
+
+
 class TestExecutionOnWindows:
     pytestmark = win32only
 
-    def test_sysfind(self):
-        x = py.path.local.sysfind('cmd')
-        assert x.check(file=1)
-        assert py.path.local.sysfind('jaksdkasldqwe') is None
-
 class TestExecution:
     pytestmark = skiponwin32
 
-    def test_sysfind(self):
-        x = py.path.local.sysfind('test')
-        assert x.check(file=1)
-        assert py.path.local.sysfind('jaksdkasldqwe') is None
-
     def test_sysfind_no_permisson_ignored(self, monkeypatch, tmpdir):
         noperm = tmpdir.ensure('noperm', dir=True)
         monkeypatch.setenv("PATH", noperm, prepend=":")
@@ -280,6 +350,14 @@ class TestExecution:
             if i>=3:
                 assert not numdir.new(ext=str(i-3)).check()
 
+    def test_make_numbered_dir_NotImplemented_Error(self, tmpdir, monkeypatch):
+        def notimpl(x, y):
+            raise NotImplementedError(42)
+        monkeypatch.setattr(py.std.os, 'symlink', notimpl)
+        x = tmpdir.make_numbered_dir(rootdir=tmpdir, lock_timeout=0)
+        assert x.relto(tmpdir)
+        assert x.check()
+
     def test_locked_make_numbered_dir(self, tmpdir):
         for i in range(10):
             numdir = local.make_numbered_dir(prefix='base2.', rootdir=tmpdir,
@@ -304,6 +382,18 @@ class TestImport:
         assert obj.x == 42
         assert obj.__name__ == 'execfile'
 
+    def test_pyimport_renamed_dir_creates_mismatch(self, tmpdir):
+        p = tmpdir.ensure("a", "test_x123.py")
+        p.pyimport()
+        tmpdir.join("a").move(tmpdir.join("b"))
+        pytest.raises(tmpdir.ImportMismatchError,
+            lambda: tmpdir.join("b", "test_x123.py").pyimport())
+
+    def test_pyimport_messy_name(self, tmpdir):
+        # http://bitbucket.org/hpk42/py-trunk/issue/129
+        path = tmpdir.ensure('foo__init__.py')
+        obj = path.pyimport()
+
     def test_pyimport_dir(self, tmpdir):
         p = tmpdir.join("hello_123")
         p_init = p.ensure("__init__.py")
@@ -371,6 +461,16 @@ class TestImport:
         assert orig == p
         assert issubclass(pseudopath.ImportMismatchError, ImportError)
 
+    def test_issue131_pyimport_on__init__(self, tmpdir):
+        # __init__.py files may be namespace packages, and thus the
+        # __file__ of an imported module may not be ourselves
+        # see issue
+        p1 = tmpdir.ensure("proja", "__init__.py")
+        p2 = tmpdir.ensure("sub", "proja", "__init__.py")
+        m1 = p1.pyimport()
+        m2 = p2.pyimport()
+        assert m1 == m2
+
 def test_pypkgdir(tmpdir):
     pkg = tmpdir.ensure('pkg1', dir=1)
     pkg.ensure("__init__.py")
@@ -397,14 +497,31 @@ def test_isimportable():
     assert not isimportable("x-1")
     assert not isimportable("x:1")
 
-def test_homedir():
+def test_homedir_from_HOME(monkeypatch):
+    path = os.getcwd()
+    monkeypatch.setenv("HOME", path)
+    assert py.path.local._gethomedir() == py.path.local(path)
+
+def test_homedir_not_exists(monkeypatch):
+    monkeypatch.delenv("HOME", raising=False)
+    monkeypatch.delenv("HOMEDRIVE", raising=False)
     homedir = py.path.local._gethomedir()
-    assert homedir.check(dir=1)
+    assert homedir is None
 
 def test_samefile(tmpdir):
     assert tmpdir.samefile(tmpdir)
     p = tmpdir.ensure("hello")
     assert p.samefile(p)
+    with p.dirpath().as_cwd():
+        assert p.samefile(p.basename)
+    if sys.platform == "win32":
+        p1 = p.__class__(str(p).lower())
+        p2 = p.__class__(str(p).upper())
+        assert p1.samefile(p2)
+
+def test_listdir_single_arg(tmpdir):
+    tmpdir.ensure("hello")
+    assert tmpdir.listdir("hello")[0].basename == "hello"
 
 def test_mkdtemp_rootdir(tmpdir):
     dtmp = local.mkdtemp(rootdir=tmpdir)
@@ -452,12 +569,9 @@ class TestWINLocalPath:
     def test_sysfind_in_currentdir(self, path1):
         cmd = py.path.local.sysfind('cmd')
         root = cmd.new(dirname='', basename='') # c:\ in most installations
-        old = root.chdir()
-        try:
+        with root.as_cwd():
             x = py.path.local.sysfind(cmd.relto(root))
             assert x.check(file=1)
-        finally:
-            old.chdir()
 
 class TestPOSIXLocalPath:
     pytestmark = skiponwin32
@@ -484,6 +598,7 @@ class TestPOSIXLocalPath:
         linkpath.mksymlinkto(filepath)
         assert linkpath.check(file=1)
         assert not linkpath.check(link=0, file=1)
+        assert linkpath.islink()
 
     def test_symlink_relative(self, tmpdir):
         linkpath = tmpdir.join('test')
@@ -545,6 +660,21 @@ class TestPOSIXLocalPath:
         assert gid == stat.gid
         assert group == stat.group
 
+    def test_stat_helpers(self, tmpdir, monkeypatch):
+        path1 = tmpdir.ensure("file")
+        stat1 = path1.stat()
+        stat2 = tmpdir.stat()
+        assert stat1.isfile()
+        assert stat2.isdir()
+        assert not stat1.islink()
+        assert not stat2.islink()
+
+    def test_stat_non_raising(self, tmpdir):
+        path1 = tmpdir.join("file")
+        pytest.raises(py.error.ENOENT, lambda: path1.stat())
+        res = path1.stat(raising=False)
+        assert res is None
+
     def test_atime(self, tmpdir):
         import time
         path = tmpdir.ensure('samplefile')
@@ -577,7 +707,7 @@ class TestPOSIXLocalPath:
     def test_join_to_root(self, path1):
         root = path1.parts()[0]
         assert len(str(root)) == 1
-        assert str(root.join('a')) == '/a'
+        assert str(root.join('a')) == '//a'  # posix allows two slashes
 
     def test_join_root_to_root_with_no_abs(self, path1):
         nroot = path1.join('/')
@@ -607,6 +737,16 @@ class TestPOSIXLocalPath:
             for x,y in oldmodes.items():
                 x.chmod(y)
 
+    def test_copy_archiving(self, tmpdir):
+        f = tmpdir.ensure("a", "file1")
+        a = f.dirpath()
+        oldmode = f.stat().mode
+        newmode = oldmode ^ 1
+        f.chmod(newmode)
+        b = tmpdir.join("b")
+        a.copy(b, mode=True)
+        assert b.join(f.basename).stat().mode == newmode
+
     @failsonjython
     def test_chown_identity(self, path1):
         owner = path1.stat().owner
@@ -629,3 +769,27 @@ class TestPOSIXLocalPath:
         owner = path1.stat().owner
         group = path1.stat().group
         path1.chown(owner, group)
+
+
+class TestUnicodePy2Py3:
+    def test_join_ensure(self, tmpdir):
+        x = py.path.local(tmpdir.strpath)
+        part = py.builtin._totext("hällo", "utf8")
+        y = x.ensure(part)
+        assert x.join(part) == y
+
+    def test_listdir(self, tmpdir):
+        x = py.path.local(tmpdir.strpath)
+        part = py.builtin._totext("hällo", "utf8")
+        y = x.ensure(part)
+        assert x.listdir(part)[0] == y
+
+    @pytest.mark.xfail(reason="changing read/write might break existing usages")
+    def test_read_write(self, tmpdir):
+        x = tmpdir.join("hello")
+        part = py.builtin._totext("hällo", "utf8")
+        x.write(part)
+        assert x.read() == part
+        x.write(part.encode(sys.getdefaultencoding()))
+        assert x.read() == part.encode(sys.getdefaultencoding())
+
