@@ -8,6 +8,14 @@ def test_enumerate():
     for i,x in enumerate(l):
         assert i == x
 
+def test_any():
+    assert not py.builtin.any([0,False, None])
+    assert py.builtin.any([0,False, None,1])
+
+def test_all():
+    assert not py.builtin.all([True, 1, False])
+    assert py.builtin.all([True, 1, object])
+
 def test_BaseException():
     assert issubclass(IndexError, py.builtin.BaseException)
     assert issubclass(Exception, py.builtin.BaseException)
@@ -125,6 +133,23 @@ def test_callable():
 def test_totext():
     py.builtin._totext("hello", "UTF-8")
 
+def test_bytes_text():
+    if sys.version_info[0] < 3:
+        assert py.builtin.text == unicode
+        assert py.builtin.bytes == str
+    else:
+        assert py.builtin.text == str
+        assert py.builtin.bytes == bytes
+
+def test_totext_badutf8():
+    # this was in printouts within the pytest testsuite
+    # totext would fail
+    if sys.version_info >= (3,):
+        errors = 'surrogateescape'
+    else: # old python has crappy error handlers
+        errors = 'replace'
+    py.builtin._totext("\xa6", "UTF-8", errors)
+
 def test_reraise():
     from py.builtin import _reraise
     try:
